/******************************************************************************
 * The MIT License (MIT)
 *
 * Copyright (c) 2019-2021 Baldur Karlsson
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 ******************************************************************************/

#pragma once
#include <QLabel>
#include <QVariant>

class RDLabel : public QLabel
{
private:
  Q_OBJECT
public:
  explicit RDLabel(QWidget *parent = 0);
  ~RDLabel();

  QSize sizeHint() const override;
  QSize minimumSizeHint() const override;

  void setText(const QString &text);
  QString text() const;

  void setMinimumSizeHint(const QSize &sz);
  void setPreserveAspectRatio(bool preserve) { m_preserveRatio = preserve; }
  bool preserveAspectRatio() { return m_preserveRatio; }
signals:
  void clicked(QMouseEvent *event);
  void doubleClicked(QMouseEvent *event);
  void mouseMoved(QMouseEvent *event);
  void leave();
  void styleChanged(QEvent *event);

public slots:

protected:
  void mousePressEvent(QMouseEvent *event) override;
  void mouseReleaseEvent(QMouseEvent *event) override;
  void mouseMoveEvent(QMouseEvent *event) override;
  void mouseDoubleClickEvent(QMouseEvent *event) override;
  void leaveEvent(QEvent *event) override;
  void resizeEvent(QResizeEvent *event) override;
  void changeEvent(QEvent *event) override;
  void paintEvent(QPaintEvent *event) override;

  void modifySizeHint(QSize &sz) const;

  bool m_preserveRatio = false;
  bool m_hover = false;

  QSize m_minSizeHint;

  QVariant m_variant;
};
