/******************************************************************************
 * The MIT License (MIT)
 *
 * Copyright (c) 2019-2021 Baldur Karlsson
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 ******************************************************************************/

#pragma once

#include <QDialog>

namespace Ui
{
class UpdateDialog;
}

class QNetworkAccessManager;
class QNetworkReply;
class QElapsedTimer;

struct Thumbnail;

class UpdateDialog : public QDialog
{
  Q_OBJECT
public:
  explicit UpdateDialog(QString updateResponse, QWidget *parent = 0);
  ~UpdateDialog();

private slots:
  // automatic slots
  void on_releaseNotes_clicked();
  void on_close_clicked();
  void on_update_clicked();

private:
  void keyPressEvent(QKeyEvent *e) override;
  void closeEvent(QCloseEvent *) override;

  QString m_NewVer;
  QString m_URL;
  uint32_t m_Size = 0;

  Ui::UpdateDialog *ui;

  QElapsedTimer *m_DownloadTimer = NULL;

  QNetworkAccessManager *m_NetManager;
  QNetworkReply *m_Request = NULL;
};
